import datetime
import io
import json
import uuid

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding

class Transaction():
    def __init__(self, transactionId=None, timestamp=None, cardId=None, sender=None,
                 receiver=None, authorPrivateKey=None, signature=None, data=None):
        self.transactionId = transactionId if transactionId else uuid.uuid4()
        self.timestamp = timestamp if timestamp else datetime.datetime.now(
            datetime.timezone.utc
            )
        self.cardId = cardId
        self.sender = sender
        self.receiver = receiver
        self.signature = signature
        if data:
            self.load_from_data(data)
        if authorPrivateKey:
            self.sign(authorPrivateKey)

    def validate(self):
        # TODO: validate transactionId
        # TODO: validate timestamp
        # TODO: validate cardId
        # TODO: validate sender
        # TODO: validate receiver
        try:
            self.sender.verify(
                self.signature,
                bytearray(
                    json.dumps({
                        "transactionId": str(self.transactionId),
                        "timestamp": str(self.timestamp),
                        "cardId": str(self.cardId),
                        "receiver": self.receiver.public_bytes(
                            encoding=serialization.Encoding.PEM,
                            format=serialization.PublicFormat.SubjectPublicKeyInfo
                            ).decode('utf-8')
                        }, sort_keys=True).encode('utf-8')
                    ),
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH
                    ),
                hashes.SHA256()
                )
        except Exception as e:
            raise self.InvalidSignature(str(e))

    def sign(self, authorPrivateKey):
        self.signature = authorPrivateKey.sign(
            bytearray(
                json.dumps({
                    "transactionId": str(self.transactionId),
                    "timestamp": str(self.timestamp),
                    "cardId": str(self.cardId),
                    "receiver": self.receiver.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo
                        ).decode('utf-8')
                    }, sort_keys=True).encode('utf-8')
                ),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
                ),
            hashes.SHA256()
            )

    def load_from_data(self, data):
        self.transactionId = uuid.UUID(data['transactionId'])
        self.timestamp = datetime.datetime.strptime(
            data['timestamp'],
            "%Y-%m-%d %H:%M:%S.%f%z"
            )
        self.cardId = uuid.UUID(data['cardId'])
        # TODO: why is this sometimes a tuple?
        self.sender = serialization.load_pem_public_key(
            data['sender'].encode('utf-8'),
            backend=default_backend()
            ),
        if isinstance(self.sender, tuple):
            self.sender = self.sender[0]
        self.receiver = serialization.load_pem_public_key(
            data['receiver'].encode('utf-8'),
            backend=default_backend()
            ),
        if isinstance(self.receiver, tuple):
            self.receiver = self.receiver[0]
        self.signature = bytes.fromhex(data['signature'])

    def as_dict(self):
        return {
            "transactionId": str(self.transactionId),
            "timestamp": str(self.timestamp),
            "cardId": str(self.cardId),
            "sender": self.sender.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
                ).decode('utf-8'),
            "receiver": self.receiver.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
                ).decode('utf-8'),
            "signature": self.signature.hex()
            }

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True)

    class Unauthorized(Exception):
        def __init__(self, message, statusCode=403):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class Invalid(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class AlreadyFulfilled(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class Abandoned(Exception):
        def __init__(self, message, statusCode=500):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class InvalidSignature(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode
